#include <ruby.h>
#include <node.h>
#include <env.h>
#include <st.h>

#include "yarvcore.h"
#include "vm.h"
#include "insnhelper.h"
#include "vm_macro.inc"
#include "insns.inc"

#define PROCDEBUG 0

#define BUFSIZE 0x100

#define EVALBODY_HELPER_FUNCTION static inline

#ifndef HAVE_TYPE_ULONG
typedef unsigned long ulong;
#endif

typedef ulong GENTRY;

void vm_analysis_operand(int insn, int n, VALUE op);
void vm_analysis_register(int reg, int isset);
void vm_analysis_insn(int insn);

static inline VALUE
th_invoke_yield_cfunc(yarv_thread_t *th,
                      yarv_block_t *block, int argc, VALUE *argv);

VALUE th_eval_body(yarv_thread_t *th);

#if OPT_STACK_CACHING
static VALUE yarv_finish_insn_seq[1] = {BIN(finish_SC_ax_ax)};
#elif OPT_CALL_THREADED_CODE
static VALUE const yarv_finish_insn_seq[1] = {0};
#else
static VALUE yarv_finish_insn_seq[1] = {BIN(finish)};
#endif

/* imporeted from eval.c */
struct cache_entry {		/* method hash table. */
    ID mid;			/* method's id */
    ID mid0;			/* method's original id */
    VALUE klass;		/* receiver's class */
    VALUE origin;		/* where method defined  */
    NODE *method;
    int noex;
};

#define CACHE_SIZE 0x800

//RUBY_EXTERN struct cache_entry ruby_cache_table[CACHE_SIZE];
//static struct cache_entry *global_method_cache = &ruby_cache_table[0];

#define cache global_method_cache

#define CACHE_MASK 0x7ff
#define EXPR1(c,m) ((((c)>>3)^(m))&CACHE_MASK)

NODE* rb_get_method_body(VALUE *klassp, ID *idp, int *noexp);

#include "call_cfunc.h"

/*
 * prepare stack frame
 */
static inline VALUE
th_set_env(yarv_thread_t *th, yarv_iseq_t *iseq,
           VALUE magic, VALUE self, VALUE specval,
           VALUE *pc, VALUE *sp, VALUE *lfp,
           int local_size, int argc, const VALUE *argv)
{
  VALUE *dfp;
  int i;

  for(i=0; i<argc; i++){
    *sp = *argv;
    sp++; argv++;
  }

  for(; i<local_size; i++){
    *sp = Qnil;
    sp++;
  }

  *sp = GC_GUARDED_PTR(specval);
  dfp = sp;
  if(lfp == 0){
    lfp = sp;
  }
  PUSH_CONTROL_STACK_FRAME(th,
                           pc,            // pc
                           sp+1,          // sp
                           sp+1,          // bp
                           iseq,          // iseq
                           magic,         // magic
                           self,          // self
                           lfp,           // lfp
                           dfp            // dfp
                           );

  return Qtrue;
}

EXTERN VALUE ruby_top_self;

VALUE
th_set_finish_env(yarv_thread_t *th)
{
  th_set_env(th, 0,
             FRAME_MAGIC_FINISH, Qnil, 0,
             0, th->cfp->sp, 0,
             0, 0, 0);
  th->cfp->pc = &yarv_finish_insn_seq[0];
  return Qtrue;
}

VALUE
th_set_top_stack(yarv_thread_t *th, VALUE iseqval)
{
  yarv_iseq_t *iseq;
  GetISeqVal(iseqval, iseq);

//  iseq->klass_nest_stack = rb_ary_push(rb_ary_new(), rb_cObject);

  /* for return */
  th_set_finish_env(th);

  return th_set_env(th, iseq,
                    FRAME_MAGIC_TOP, ruby_top_self, 0,
                    iseq->iseq_encoded, th->cfp->sp, 0,
                    iseq->local_size, 0, 0);
}

VALUE
th_set_eval_stack(yarv_thread_t *th, VALUE iseqval){
  yarv_iseq_t *iseq;
  yarv_block_t *block = th->base_block;
  GetISeqVal(iseqval, iseq);

  /* for return */
  th_set_finish_env(th);
  th_set_env(th, iseq,
             FRAME_MAGIC_EVAL, block->self, GC_GUARDED_PTR(block->dfp),
             iseq->iseq_encoded, th->cfp->sp, block->lfp,
             iseq->local_size, 0, 0);
  return 0;
}

static int check_env(yarv_env_t *env);

static VALUE
th_make_env_each(yarv_thread_t *th, yarv_control_frame_t *cfp,
                 VALUE *envptr, VALUE *endptr)
{
  VALUE envval, penvval = 0;
  yarv_env_t *env;
  VALUE *nenvptr;
  int i, local_size;

  if(ENV_IN_HEAP_P(envptr)){
    return ENV_VAL(envptr);
  }

  if(envptr != endptr){
    VALUE *penvptr = GC_GUARDED_PTR_REF(*envptr);
    yarv_control_frame_t *pcfp = cfp;

    if(ENV_IN_HEAP_P(penvptr)){
      penvval = ENV_VAL(penvptr);
    }
    else{
      while(pcfp->dfp != penvptr){
        // printf("pcfp: %p (%p)\n", pcfp->dfp, penvptr);
        pcfp++;
        if(pcfp->dfp == 0){
          SDR();
          printf("[BUG] orz\n");
          exit(0);
        }
      }
      penvval = th_make_env_each(th, pcfp, penvptr, endptr);
      cfp->lfp = pcfp->lfp;
      *envptr = GC_GUARDED_PTR(pcfp->dfp);
    }
  }

  /* allocate env */
  envval = rb_obj_alloc(cYarvEnv);
  GetEnvVal(envval, env);

  if(cfp->iseq == 0 || CMETHOD_INFO_P(cfp->iseq)){
    local_size = 2;
  }
  else{
    local_size = cfp->iseq->local_size;
  }

  env->env_size    = local_size + 1 + 4;
  env->local_size  = local_size;
  env->env         = ALLOC_N(VALUE, env->env_size);
  env->prev_envval = penvval;

  for(i=0; i<=local_size; i++){
    env->env[i] = envptr[-local_size + i];
    if(cfp->iseq != 0 && !CMETHOD_INFO_P(cfp->iseq)){
      /* clear value stack for GC */
      // envptr[-local_size + i] = 0;
    }
  }

  *envptr = envval; /* GC mark */
  nenvptr = &env->env[i-1];
  nenvptr[1] = Qfalse; /* frame is not orphan */
  nenvptr[2] = Qundef; /* frame is in heap    */
  nenvptr[3] = envval; /* frame self */
  nenvptr[4] = penvval;/* frame prev env object */

  /* reset lfp/dfp in cfp */
  cfp->dfp = nenvptr;
  if(envptr == endptr){
    cfp->lfp = nenvptr;
  }

  /* as Binding */
  env->block.self = cfp->self;
  env->block.lfp  = cfp->lfp;
  env->block.dfp  = cfp->dfp;
  env->block.iseq = cfp->iseq;

  if(cfp->iseq && CMETHOD_INFO_P(cfp->iseq)){
    /* TODO */
    env->block.iseq = 0;
  }
  return envval;
}

static VALUE
check_env_value(VALUE envval);

static int
check_env(yarv_env_t *env){
  printf("---\n");
  printf("envptr: %p\n", &env->block.dfp[0]);
  printf("orphan: %p\n", (void*)env->block.dfp[1]);
  printf("inheap: %p\n", (void*)env->block.dfp[2]);
  printf("envval: %10p ", (void*)env->block.dfp[3]); dp(env->block.dfp[3]);
  printf("penvv : %10p ", (void*)env->block.dfp[4]); dp(env->block.dfp[4]);
  printf("lfp:    %10p\n", env->block.lfp);
  printf("dfp:    %10p\n", env->block.dfp);
  if(env->block.dfp[4]){
    printf(">>\n");
    check_env_value(env->block.dfp[4]);
    printf("<<\n");
  }
  return 1;
}

static VALUE
check_env_value(VALUE envval){
  yarv_env_t *env;
  GetEnvVal(envval, env);
  
  if(check_env(env)){
    return envval;
  }
  rb_bug("invalid env\n");
  return Qnil; /* unreachable */
}

static int
collect_local_variables_in_env(yarv_env_t *env, VALUE ary){
  int i;
  if(env->block.lfp == env->block.dfp){
    return 0;
  }
  for(i = 0; i<env->block.iseq->local_size; i++){
    rb_ary_push(ary, rb_str_new2(rb_id2name(env->block.iseq->local_tbl[i])));
  }
  if(env->prev_envval){
    GetEnvVal(env->prev_envval, env);
    collect_local_variables_in_env(env, ary);
  }
  return 0;
}

int
th_collect_local_variables_in_heap(yarv_thread_t *th, VALUE *dfp, VALUE ary)
{
  if(ENV_IN_HEAP_P(dfp)){
    yarv_env_t *env;
    GetEnvVal(ENV_VAL(dfp), env);
    collect_local_variables_in_env(env, ary);
    return 1;
  }
  else{
    return 0;
  }
}


VALUE
th_make_env_object(yarv_thread_t *th, yarv_control_frame_t *cfp)
{
  VALUE envval;
  envval = th_make_env_each(th, cfp, cfp->dfp, cfp->lfp);
  if(PROCDEBUG){
    check_env_value(envval);
  }
  return envval;
}

static VALUE
th_make_proc_from_block(yarv_thread_t *th, yarv_control_frame_t *cfp,
                        yarv_block_t *block)
{
  VALUE procval;
  yarv_control_frame_t *bcfp;
  VALUE *bdfp; /* to gc mark */

  if(block->proc){
    return block->proc;
  }
  
  bcfp = GET_CFP_FROM_BLOCK_PTR(block);
  bdfp = bcfp->dfp;
  procval = th_make_proc(th, bcfp, block);
  /* what is this? */
  // bdfp[-1] = procval;
  return procval;
}

struct RObject *rb;

VALUE
th_make_proc(yarv_thread_t *th, yarv_control_frame_t *cfp, yarv_block_t *block)
{
  VALUE procval, envval, blockprocval = 0;
  yarv_proc_t *proc;

  if(GC_GUARDED_PTR_REF(cfp->lfp[0]) != 0 &&
     !(cfp->lfp[0] & 0x02) /* not class scope */){
    yarv_proc_t *p;

    blockprocval =
      th_make_proc_from_block(th, cfp, (yarv_block_t *)GC_GUARDED_PTR_REF(*cfp->lfp));
    GetProcVal(blockprocval, p);
    *cfp->lfp = GC_GUARDED_PTR(&p->block);
  }
  envval = th_make_env_object(th, cfp);

  if(PROCDEBUG){
    check_env_value(envval);
  }

  procval = rb_obj_alloc(cYarvProc);
  GetProcVal(procval, proc);

  proc->blockprocval = blockprocval;
  proc->block.self   = block->self;
  proc->block.lfp    = block->lfp;
  proc->block.dfp    = block->dfp;
  proc->block.iseq   = block->iseq;
  proc->block.proc   = procval;
  proc->envval       = envval;

  /* set cref */
  if(BUILTIN_TYPE(block->iseq) != T_NODE){
    if(cfp->iseq && cfp->iseq->klass_nest_stack){
      block->iseq->klass_nest_stack = cfp->iseq->klass_nest_stack;
    }
    else{
      block->iseq->klass_nest_stack = rb_ary_dup(th->klass_nest_stack);
    }
  }
  return procval;
}

VALUE
th_call0(yarv_thread_t *th, VALUE klass, VALUE recv,
         VALUE id, ID oid, int argc, const VALUE *argv,
         NODE *body, int nosuper)
{
  VALUE val;
  yarv_block_t *blockptr = 0;

  //printf("id: %s, argc: %d\n", rb_id2name(id), argc);
  //SDR2(th->cfp);
  
  if(th->passed_block){
    blockptr = th->passed_block;
    th->passed_block = 0;
  }
  
  switch(nd_type(body)){
  case YARV_METHOD_NODE:{
    yarv_control_frame_t *reg_cfp;
    int i;
    const int flag = 0;
    
    th_set_finish_env(th);
    reg_cfp = th->cfp;
    for(i=0; i<argc; i++){
      *reg_cfp->sp++ = argv[i];
    }
    macro_eval_invoke_func(body->nd_body, recv, klass, blockptr, argc);
    val = th_eval_body(th);
    break;
  }
  case NODE_CFUNC:{
    yarv_control_frame_t *reg_cfp = th->cfp;
    struct yarv_cmethod_info cmi = {0, id, klass};

    th_set_env(th, (yarv_iseq_t *)&cmi,
               FRAME_MAGIC_CFUNC, recv, (VALUE)blockptr,
               0, reg_cfp->sp, 0,
               0, 0, 0);
    val = call_cfunc(body->nd_cfnc, recv, body->nd_argc, argc, argv);
    th->cfp = reg_cfp; /* pop control stack frame */
    break;
  }
  case NODE_ATTRSET:{
    if (argc != 1){
      rb_raise(rb_eArgError, "wrong number of arguments (%d for 1)", argc);
    }
    val = rb_ivar_set(recv, body->nd_vid, argv[0]);
    break;
  }
  case NODE_IVAR:{
    if (argc != 0) {
      rb_raise(rb_eArgError, "wrong number of arguments (%d for 0)", argc);
    }
    val = rb_attr_get(recv, body->nd_vid);
    break;
  }
  case NODE_BMETHOD:{
    yarv_control_frame_t *reg_cfp;
    int i;
    const int flag = 0;
    
    th_set_finish_env(th);
    
    reg_cfp = th->cfp;
    
    for(i=0; i<argc; i++){
      *reg_cfp->sp++ = argv[i];
    }
    macro_eval_invoke_bmethod(body->nd_cval, recv, klass, blockptr, argc);
    val = th_eval_body(th);
    break;
  }
  default:
    rb_bug("unsupported: thread_call0");
  }
  return val;
}

VALUE
thread_call0(VALUE self, VALUE klass, VALUE recv, VALUE id, ID oid,
             int argc, const VALUE *argv, NODE *body, int nosuper)
{
  yarv_thread_t *th;
  GetThreadVal(self, th);
  return th_call0(th, klass, recv, id, oid, argc, argv, body, nosuper);
}

VALUE
th_call_super(yarv_thread_t *th, int argc, const VALUE *argv)
{
  VALUE recv  = th->cfp->self;
  VALUE klass;
  ID id;
  NODE *body;
  int nosuper = 0;
  {
    struct yarv_cmethod_info *cmi = (void *)th->cfp->iseq;
    if(cmi->sig == 0){
      klass = RCLASS(cmi->klass)->super;
      id = cmi->id;
    }
    else{
      rb_bug("thread_call_super: should not be reached");
    }
  }
  
  body = rb_method_node(klass, id);
  return th_call0(th, klass, recv, id, id, argc, argv, body, nosuper);
}


VALUE
thread_call_super(VALUE self, int argc, const VALUE *argv)
{
  yarv_thread_t *th;
  GetThreadVal(self, th);
  return th_call_super(th, argc, argv);
}


static inline VALUE
th_invoke_yield_cfunc(yarv_thread_t *th,
                      yarv_block_t *block, int argc, VALUE *argv)
{
  NODE *ifunc = (NODE*)block->iseq;
  VALUE val;
  VALUE arg;

  if(argc == 1){
    arg = *argv;
  }
  else if(argc > 1){
    arg = rb_ary_new4(argc, argv);
  }
  else{
    arg = rb_ary_new();
  }

  th_set_env(th, 0,
             FRAME_MAGIC_IFUNC, block->self, (VALUE)block->dfp,
             0, th->cfp->sp, block->lfp,
             0, 0, 0);

  val = (*ifunc->nd_cfnc)(arg, ifunc->nd_tval, Qnil);
  
  th->cfp++;
  return val;
}


VALUE
th_invoke_yield(yarv_thread_t *th, int argc, VALUE *argv)
{
  yarv_control_frame_t *cfp = th->cfp;
  yarv_block_t *block = GC_GUARDED_PTR_REF(cfp->lfp[0]);
  VALUE val;
  if(block == 0){
    rb_bug("block is not given (th_invoke_yield)");
  }
  else{
    if(BUILTIN_TYPE(block->iseq) != T_NODE){

      if(block->iseq->argc == 1){
        if(argc > 1){
          argv[0] = rb_ary_new4(argc, argv);
          argc = 1;
        }
      }
      else{
        if(argc == 1 && !SPECIAL_CONST_P(argv[0]) &&
           BUILTIN_TYPE(argv[0]) == T_ARRAY){
          argc = RARRAY(argv[0])->len;
          argv = RARRAY(argv[0])->ptr;
        }
      }

      if(block->iseq->arg_rest != 0){
        int rest = block->iseq->arg_rest - 1;
        int alen;
        
        if(argc <= rest){
          VALUE ary = rb_ary_new4(argc, argv);
          rb_ary_store(ary, rest, rb_ary_new());
          argc = RARRAY(ary)->len;
          argv = RARRAY(ary)->ptr;
        }
        else{
          argv[rest] = rb_ary_new4(argc-rest, &argv[rest]);
          argc = rest + 1;
        }
      }

      th_set_finish_env(th);
      th_set_env(th, block->iseq, FRAME_MAGIC_BLOCK,
                 block->self, GC_GUARDED_PTR(block->dfp),
                 block->iseq->iseq_encoded, th->cfp->sp, block->lfp,
                 block->iseq->local_size, argc, argv);
      val = th_eval_body(th);
    }
    else{
      val = th_invoke_yield_cfunc(th, block, argc, argv);
    }
  }
  return val;
}

static VALUE
th_invoke_proc_under(yarv_thread_t *th, yarv_proc_t *proc,
                     int argc, VALUE *argv, VALUE self)
{
  if(BUILTIN_TYPE(proc->block.iseq) == T_NODE){
    return th_invoke_yield_cfunc(th, &proc->block, argc, argv);
  }
  else{
    th_set_finish_env(th);
    th_set_env(th, proc->block.iseq,
               FRAME_MAGIC_PROC, self, (VALUE)proc->block.dfp,
               proc->block.iseq->iseq_encoded, th->cfp->sp, proc->block.lfp,
               proc->block.iseq->local_size, argc, argv);
  }
  return th_eval_body(th);
}

VALUE
th_invoke_proc(yarv_thread_t *th, yarv_proc_t *proc, int argc, VALUE *argv)
{
  if(BUILTIN_TYPE(proc->block.iseq) == T_NODE){
    return th_invoke_yield_cfunc(th, &proc->block, argc, argv);
  }
  else{
    return th_invoke_proc_under(th, proc, argc, argv, proc->block.self);
  }
}

VALUE
thread_invoke_proc_call(VALUE self, VALUE procval, int argc, VALUE *argv)
{
  yarv_thread_t *th;
  yarv_proc_t *proc;
  GetThreadVal(self, th);
  GetProcVal(procval, proc);
  
  return th_invoke_proc(th, proc, argc, argv);
}

VALUE *
th_svar(yarv_thread_t *th, int cnt)
{
  yarv_control_frame_t *cfp = th->cfp;
  NODE *node;

  while(cfp->pc == 0){
    cfp++;
  }
  
  node = (NODE*)cfp->lfp[-1];

  if((VALUE)node == Qnil){
    cfp->lfp[-1] = (VALUE)(node = NEW_IF(Qnil, Qnil, 0));
  }
  // printf("== %p, %p\n", node, &cfp->lfp[-1]);
  
  if(cnt == 0){
    return &node->u1.value;
  }
  else if(cnt == 1){
    return &node->u2.value;
  }
  else{
    VALUE ary;
    if((ary = node->u3.value) == 0){
      ary = node->u3.value = rb_ary_new();
    }
    if(RARRAY(ary)->len <= cnt){
      rb_ary_store(ary, cnt, Qnil);
    }
    return &RARRAY(ary)->ptr[cnt];
  }
}

VALUE *
thread_svar(VALUE self, int cnt)
{
  yarv_thread_t *th;
  GetThreadVal(self, th);
  return th_svar(th, cnt);
}

int
th_get_sourceline(yarv_control_frame_t *cfp)
{
  int i;
  int pos = cfp->pc - cfp->iseq->iseq_encoded;
  yarv_iseq_t *iseq = cfp->iseq;
  int line_no;

  for(i=0; i<iseq->insn_info_size; i++){
    if(iseq->insn_info_tbl[i].position == pos){
      line_no = iseq->insn_info_tbl[i-1].line_no;
      goto found;
    }
  }
  line_no = iseq->insn_info_tbl[i-1].line_no;

found:
  return line_no;
}

static VALUE
th_backtrace_each(yarv_thread_t *th,
                        yarv_control_frame_t *limit_cfp,
                        yarv_control_frame_t *cfp,
                        char *file, int line_no, VALUE ary)
{
  VALUE str = 0;
  char buf[BUFSIZE];

  if(cfp->iseq != 0){
    if(cfp->pc != 0){
      line_no = th_get_sourceline(cfp);
      snprintf(buf, BUFSIZ, "%s:%d:in `%s'",
               file = RSTRING(cfp->iseq->file_name)->ptr,
               line_no, RSTRING(cfp->iseq->name)->ptr);
    }
    else{
      struct yarv_cmethod_info *cmi = (void *)cfp->iseq;
      snprintf(buf, BUFSIZE, "%s:%d:in `%s'",
               file, line_no, rb_id2name(cmi->id));
    }
    str = rb_str_new2(buf);
  }
  else if(0 && cfp->magic == FRAME_MAGIC_IFUNC){
    snprintf(buf, BUFSIZE, "::in `<ifunc>'"); /*, ""%s:%d:in `<ifunc>'",
        file, line_no);*/
    str = rb_str_new2(buf);
  }
  
  if(limit_cfp != cfp){
    th_backtrace_each(th, limit_cfp, cfp-1, file, line_no, ary);
  }
  if(str){
    rb_ary_push(ary, str);
  }
  return str;
}

VALUE
th_backtrace(yarv_thread_t *th, int lev)
{
  VALUE ary;
  yarv_control_frame_t *cfp = th->cfp;
  yarv_control_frame_t *top_of_cfp = (void *)(th->stack + th->stack_size);
  top_of_cfp--;

  if(lev < 0){
    /* TODO ?? */
    ary = rb_ary_new();
  }
  else{
    while(lev-->=0){
      cfp++;
      if(cfp >= top_of_cfp){
        return Qnil;
      }
    }
    ary = rb_ary_new();
  }
  
  th_backtrace_each(th, cfp, top_of_cfp, "", 0, ary);
  return ary;
}

VALUE
thread_backtrace(VALUE self, int level)
{
  yarv_thread_t *th;
  GetThreadVal(self, th);
  return th_backtrace(th, level);
}

/*
 *
 * vm main loop helper functions
 *
 */


EVALBODY_HELPER_FUNCTION
VALUE
eval_get_ev_const(yarv_thread_t *th,
                  yarv_iseq_t *iseq, VALUE klass, ID id)
{
  VALUE val;
  
  if(klass == Qnil){
    /* in current lexical scope */
    VALUE cref; /* Array */
    int i, last_pos;

    cref = iseq->klass_nest_stack;
    if(cref == 0){
      /* in module or class scope */
      cref = th->klass_nest_stack;
    }
    
    last_pos = RARRAY(cref)->len-1;
    for(i = last_pos; i>0; i--){
      klass = RARRAY(cref)->ptr[i];
    search_continue:
      if(RCLASS(klass)->iv_tbl &&
         st_lookup(RCLASS(klass)->iv_tbl, id, &val)){
        if(val == Qundef){
          rb_autoload_load(klass, id);
          goto search_continue;
        }
        else{
          break;
        }
      }
    }
    if(i==0){
      klass = RARRAY(cref)->ptr[last_pos];
      val = rb_const_get(klass, id);
    }
  }
  else{
    val = rb_const_get(klass, id);
  }
  return val;
}

EVALBODY_HELPER_FUNCTION int
eval_get_ev_defined(yarv_thread_t *th,
                    yarv_iseq_t *iseq, VALUE klass, ID id)
{
  VALUE val;
  if(klass == Qnil){
    VALUE cref;
    int i, last_pos;

    if((cref = iseq->klass_nest_stack) == 0){
      cref = th->klass_nest_stack;
    }
    last_pos = RARRAY(cref)->len-1;
    for(i = last_pos; i>0; i--){
      klass = RARRAY(cref)->ptr[i];
    search_continue:
      if(RCLASS(klass)->iv_tbl &&
         st_lookup(RCLASS(klass)->iv_tbl, id, &val)){
        if(val == Qundef){
          rb_autoload_load(klass, id);
          goto search_continue;
        }
        else{
          return 1;
        }
      }
    }
    klass = RARRAY(cref)->ptr[last_pos];
    return rb_const_defined(klass, id);
  }
  else{
    return rb_const_defined(klass, id);
  }
}

EVALBODY_HELPER_FUNCTION VALUE
eval_search_super_klass(VALUE klass, VALUE recv)
{
  if(BUILTIN_TYPE(klass) == T_CLASS){
    klass = RCLASS(klass)->super;
  }
  else if(BUILTIN_TYPE(klass) == T_MODULE){
    VALUE k = CLASS_OF(recv);
    while(k){
      if(BUILTIN_TYPE(k) == T_ICLASS &&
         RBASIC(k)->klass == klass){
        klass = RCLASS(k)->super;
        break;
      }
      k = RCLASS(k)->super;
    }
  }
  return klass;
}

EVALBODY_HELPER_FUNCTION VALUE
eval_method_missing(yarv_thread_t *th, ID id, VALUE recv, int num, int opt){
  yarv_control_frame_t *reg_cfp = th->cfp;
  VALUE *argv = STACK_ADDR_FROM_TOP(num+1);
  VALUE val;
  argv[0] = ID2SYM(id);
  val = rb_funcall2(recv, idMethodMissing, num+1, argv);
  POPN(num+1);
  return val;
}

EVALBODY_HELPER_FUNCTION NODE*
eval_method_search(VALUE id, VALUE klass, IC ic, int *noex)
{
  NODE *mn;

//#undef  INLINE_METHOD_CACHE
//#define FAKE_INLINE_METHOD_CACHE
//#define FAKE_INLINE_METHOD_CACHE2
//#define INLINE_METHOD_CACHE
//#define ONLY_GLOBAL_METHOD_CACHE

#if 0
1  error
#elif 0 // defined(FAKE_INLINE_METHOD_CACHE)
  {
    /* don't support polymorphic method call */
    if(LIKELY(ic->ic_method != 0)){
      mn = ic->ic_method;
    }
    else{
      ic->ic_method = mn = rb_method_node(klass, id);
    }
  }
#elif 0 //defined(FAKE_INLINE_METHOD_CACHE2)
  {
    if(LIKELY(klass == ic->ic_klass)){
      mn = ic->ic_method;
    }
    else{
      mn = rb_method_node(klass, id);
      ic->ic_klass  = klass;
      ic->ic_method = mn;
    }
  }
#elif OPT_INLINE_METHOD_CACHE
  {
    if(LIKELY(klass == ic->ic_klass) &&
       LIKELY(GET_VM_STATE_VERSION() == ic->ic_vmstat)){
      mn = ic->ic_method;
    }
    else{
      mn = rb_method_node(klass, id);
      ic->ic_klass  = klass;
      ic->ic_method = mn;
      ic->ic_vmstat = GET_VM_STATE_VERSION();
    }
  }
#elif 0
  {
    struct cache_entry *ent = cache + EXPR1(klass, id);
    int noex;
    
    if (LIKELY(ent->mid == id) && LIKELY(ent->klass == klass)){
      if(LIKELY(ent->method != 0)){
        // klass = ent->origin;
        // id    = ent->mid0;
        // noex  = ent->noex;
        mn    = ent->method;
      }
      else{
        mn = 0;
      }
    }
    else{
      mn = rb_method_node(klass, id);
    }
  }
#else
  mn = rb_method_node(klass, id);
#endif
  return mn;
}

static void
call_yarv_end_proc(VALUE data)
{
  rb_proc_call(data, rb_ary_new2(0));
}


/*********************************************************/
/*********************************************************/
/* import from ruby1.9/eval.c                            */

#include "eval_intern.h"

/*********************************************************/
/*********************************************************/

NORETURN(static void localjump_error(const char*, VALUE, int));
static void
localjump_error(mesg, value, reason)
    const char *mesg;
    VALUE value;
    int reason;
{
  VALUE exc = rb_exc_new2(rb_const_get(rb_cObject, rb_intern("LocalJumpError")), mesg);
  ID id;

  rb_iv_set(exc, "@exit_value", value);
  switch (reason) {
  case TAG_BREAK:
    id = rb_intern("break"); break;
  case TAG_REDO:
    id = rb_intern("redo"); break;
  case TAG_RETRY:
    id = rb_intern("retry"); break;
  case TAG_NEXT:
    id = rb_intern("next"); break;
  case TAG_RETURN:
    id = rb_intern("return"); break;
  default:
    id = rb_intern("noreason"); break;
  }
  rb_iv_set(exc, "@reason", ID2SYM(id));
  rb_exc_raise(exc);
}

void
th_iter_break(yarv_thread_t *th){
  yarv_control_frame_t *cfp = th->cfp;
  VALUE *dfp = GC_GUARDED_PTR_REF(*cfp->dfp);
  
  th->state = TAG_BREAK;
  th->errinfo = (VALUE)NEW_THROW_OBJECT(Qnil, (VALUE)dfp, TAG_BREAK);
  TH_JUMP_TAG(th, TAG_BREAK);
}

VALUE yarv_unredefined_flag = 0x01;

#include "vm_evalbody.h"

/*                  finish
  VMe (h1)          finish
    VM              finish F1 F2
      func          finish F1 F2 C1
        rb_funcall  finish F1 F2 C1
          VMe       finish F1 F2 C1
            VM      finish F1 F2 C1 F3

  F1 - F3 : pushed by VM
  C1      : pushed by send insn (CFUNC)

  struct CONTROL_FRAME{
    VALUE *pc;                  // cfp[0]
    VALUE *sp;                  // cfp[1]
    VALUE *bp;                  // cfp[2]
    yarv_iseq_t *iseq;          // cfp[3]
    VALUE magic;                // cfp[4]
    VALUE self;                 // cfp[5]
    VALUE *lfp;                 // cfp[6]
    VALUE *dfp;                 // cfp[7]
    yarv_iseq_t * block_iseq;   // cfp[8]
    VALUE proc;                 // cfp[9] always 0
  };

  struct BLOCK{
    VALUE self;
    VALUE *lfp;
    VALUE *dfp;
    yarv_iseq_t *block_iseq;
  };

  struct PROC{
    VALUE  proc_sig = 0;
    struct BLOCK;
  };

  struct METHOD_CONTROL_FRAME{
    struct CONTROL_FRAME;
  };

  struct METHOD_FRAME{
    VALUE arg0;
    ...
    VALUE argM;
    VALUE param0;
    ...
    VALUE paramN;
    VALUE special;                         // lfp [1]
    struct block_object *block_ptr | 0x01; // lfp [0]
  };

  struct BLOCK_CONTROL_FRAME{
    struct STACK_FRAME;
  };

  struct BLOCK_FRAME{
    VALUE arg0;
    ...
    VALUE argM;
    VALUE param0;
    ...
    VALUE paramN;
    VALUE *(prev_ptr | 0x01); // DFP[0]
  };

  struct CLASS_CONTROL_FRAME{
    struct STACK_FRAME;
  };

  struct CLASS_FRAME{
    VALUE param0;
    ...
    VALUE paramN;
    VALUE prev_dfp; // for frame jump
  };

  struct C_METHOD_CONTROL_FRAME{
    VALUE *pc;                       // 0
    VALUE *sp;                       // stack pointer
    VALUE *bp;                       // base pointer (used in exception)
    yarv_iseq_t *iseq;               // cmi
    VALUE magic;                     // C_METHOD_FRAME
    VALUE self;                      // ?
    VALUE *lfp;                      // lfp
    VALUE *dfp;                      // == lfp
    yarv_iseq_t * block_iseq;        //
    VALUE proc;                      // always 0
  };

  struct C_BLOCK_CONTROL_FRAME{
    VALUE *pc;                       // finish wƂDreturn 邾D
    VALUE *sp;                       // sp
    yarv_iseq_t *iseq;        // ?
    VALUE magic;                     // C_METHOD_FRAME
    VALUE self;                      // H@ςv邩ȁH
    VALUE *lfp;                      // lfp
    VALUE *dfp;                      // lfp
    yarv_iseq_t * block_iseq; // 0
  };
  
  struct C_METHDO_FRAME{
    VALUE block_ptr;
    VALUE special;
  };
 */


VALUE
th_eval_body(yarv_thread_t *th)
{
  struct yarv_tag _tag;
  int state;
  VALUE result, err;
  VALUE initial = 0;
  
  TH_PUSH_TAG(th);
  if((state = EXEC_TAG()) == 0){
  vm_loop_start:
    result = th_eval(th, initial);
    if((state = th->state) != 0){
      err = result;
      th->state = 0;
      goto exception_handler;
    }
  }
  else{
    int i;
    struct catch_table_entry *entry;
    ulong  epc, cont_pc, cont_sp;
    VALUE catch_iseqval;
    yarv_control_frame_t *cfp;
    VALUE  *escape_dfp;
    VALUE type;
    
    err = th->errinfo;
    if(state == TAG_RAISE){
      rb_ivar_set(err, idThrowState, state);
    }
    
  exception_handler:
    cont_pc = cont_sp = catch_iseqval = 0;
    
    while(th->cfp->pc == 0 ||
          th->cfp->iseq == 0){
      th->cfp++;
    }
    
    cfp = th->cfp;
    epc = cfp->pc - cfp->iseq->iseq_encoded;

    if(state == TAG_BREAK || state == TAG_RETURN){
      escape_dfp = GET_THROWOBJ_CATCH_POINT(err);

      if(cfp->dfp == escape_dfp){
        if(state == TAG_RETURN){
          if((cfp+1)->pc != &yarv_finish_insn_seq[0]){
            SET_THROWOBJ_CATCH_POINT(err, (VALUE)(cfp+1)->dfp);
            SET_THROWOBJ_STATE      (err, state = TAG_BREAK);
          }
          else{
            result = GET_THROWOBJ_VAL(err);
            th->errinfo = Qnil;
            goto finish_vme;
          }
          /* through */
        }
        else{
          /* TAG_BREAK */
#if OPT_STACK_CACHING
          initial = (GET_THROWOBJ_VAL(err));
#else
          *th->cfp->sp++ = (GET_THROWOBJ_VAL(err));
#endif
          th->errinfo = Qnil;
          goto vm_loop_start;
        }
      }
    }

    if(state == TAG_RAISE){
      for(i=0; i<cfp->iseq->catch_table_size; i++){
        entry = &cfp->iseq->catch_table[i];
        if(entry->start <  epc &&
           entry->end   >= epc){

          if(entry->type == CATCH_TYPE_RESCUE ||
             entry->type == CATCH_TYPE_ENSURE){
            catch_iseqval = entry->iseq;
            cont_pc    = entry->cont;
            cont_sp    = entry->sp;
            break;
          }
        }
      }
    }
    else if(state == TAG_RETRY){
      for(i=0; i<cfp->iseq->catch_table_size; i++){
        entry = &cfp->iseq->catch_table[i];
        if(entry->start <  epc &&
           entry->end   >= epc){
          
          if(entry->type == CATCH_TYPE_ENSURE){
            catch_iseqval = entry->iseq;
            cont_pc    = entry->cont;
            cont_sp    = entry->sp;
            break;
          }
          else if(entry->type == CATCH_TYPE_RETRY){
            VALUE  *escape_dfp;
            escape_dfp = GET_THROWOBJ_CATCH_POINT(err);
            if(cfp->dfp == escape_dfp){
              cfp->pc = cfp->iseq->iseq_encoded + entry->cont;
              th->errinfo = Qnil;
              goto vm_loop_start;
            }
          }
        }
      }
    }
    else if(state == TAG_BREAK &&
            ((VALUE)escape_dfp & ~0x03) == 0){
      type = CATCH_TYPE_BREAK;

    search_restart_point:
      for(i=0; i<cfp->iseq->catch_table_size; i++){
        entry = &cfp->iseq->catch_table[i];
        if(entry->start <  epc &&
           entry->end   >= epc){
          
          if(entry->type == CATCH_TYPE_ENSURE){
            catch_iseqval = entry->iseq;
            cont_pc       = entry->cont;
            cont_sp       = entry->sp;
            break;
          }
          else if(entry->type == type){
            cfp->pc = cfp->iseq->iseq_encoded + entry->cont;
            cfp->sp = cfp->bp + entry->sp;
            
            if(!(state == TAG_REDO) &&
               !(state == TAG_NEXT  && !escape_dfp) &&
               !(state == TAG_BREAK && !escape_dfp)){
#if OPT_STACK_CACHING
              initial = (GET_THROWOBJ_VAL(err));
#else
              *th->cfp->sp++ = (GET_THROWOBJ_VAL(err));
#endif
            }
            th->errinfo = Qnil;
            goto vm_loop_start;
          }
        }
      }
    }
    else if(state == TAG_REDO){
      type = CATCH_TYPE_REDO;
      escape_dfp = GET_THROWOBJ_CATCH_POINT(err);
      goto search_restart_point;
    }
    else if(state == TAG_NEXT){
      type = CATCH_TYPE_NEXT;
      escape_dfp = GET_THROWOBJ_CATCH_POINT(err);
      goto search_restart_point;
    }
    else{
      for(i=0; i<cfp->iseq->catch_table_size; i++){
        entry = &cfp->iseq->catch_table[i];
        if(entry->start <  epc &&
           entry->end   >= epc){
          
          if(entry->type == CATCH_TYPE_ENSURE){
            catch_iseqval = entry->iseq;
            cont_pc    = entry->cont;
            cont_sp    = entry->sp;
            break;
          }
        }
      }
    }

    if(catch_iseqval != 0){
      /* found catch table */
      yarv_iseq_t* catch_iseq;
      VALUE *sp;
      
      /* enter catch scope */
      GetISeqVal(catch_iseqval, catch_iseq);
      cfp->sp = cfp->bp + cont_sp;
      cfp->pc = cfp->iseq->iseq_encoded + cont_pc;

      /* push block frame */
      th_set_env(th, catch_iseq,
                 FRAME_MAGIC_BLOCK, cfp->self, GC_GUARDED_PTR(cfp->dfp),
                 catch_iseq->iseq_encoded, cfp->sp, cfp->lfp,
                 catch_iseq->local_size - 1, 1, &err);
      
      state = 0;
      th->errinfo = Qnil;
      goto vm_loop_start;
    }
    else{
      /* pop cref */
      if(cfp->iseq->type == ISEQ_TYPE_CLASS){
        rb_ary_pop(th->klass_nest_stack);
      }
      
      th->cfp++;
      if(th->cfp->pc != &yarv_finish_insn_seq[0]){
        goto exception_handler;
      }
      else{
        th->cfp++;
        th->errinfo = err;
        TH_POP_TAG2();
        JUMP_TAG(state);
      }
    }
  }
finish_vme:
  TH_POP_TAG();
  return result;
}


VALUE
thread_eval(VALUE self)
{
  yarv_thread_t *thobj;
  yarv_iseq_t *iseq;
  
  GetThreadVal(self , thobj);
  return th_eval_body(thobj);
}

